/***************************************************************************
                           kxinewidget.h -
		A KDE/QT API for the xine-lib. Define USE_QT_ONLY if you use QT only.
                             -------------------
    begin                : Fre Apr 18 2003
    revision             : $Revision: 1.21 $
    last modified        : $Date: 2005/07/11 17:52:33 $ by $Author: juergenk $
    copyright            : (C) 2003-2005 by Jürgen Kofler
    email                : kaffeine@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KXINEWIDGET_H
#define KXINEWIDGET_H

/* define this if you want to use QT only - disables post plugin support (except visualization)! */
//#define USE_QT_ONLY 1

/* Usage:
 *
 * KXineWidget m_xine = new KXineWidget(parent);
 * m_xine->appendToQueue("the_best_song_of_the_world.mp3");
 *
 * You should at least connect to signalXineFatal(const QString&)!
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* forward declaration */
class QWidget;
class QString;
class QStringList;
class QMouseEvent;
class QTimerEvent;
class QTime;

#include <qtimer.h>
#include <qthread.h>
#include <qptrlist.h>
#include <X11/Xlib.h>
#include <xine.h>

#ifndef USE_QT_ONLY
#include "postfilter.h"
#else
 typedef int PostFilter; /* dummy type */
#endif

#define SUPPORTED_PROTOCOLS "file,http,mms,mmst,rtsp,rtp,tcp,pnm,cdda,vcd,vcdo,dvd,dvb,pvr,v4l,net,vdr,smb"

#define DEFAULT_TVTIME_CONFIG "tvtime:method=LinearBlend,enabled=1,pulldown=none,framerate_mode=half_top,judder_correction=0,use_progressive_frame_flag=1,chroma_filter=0,cheap_mode=1"

#define DEFAULT_OSD_FONT_SIZE 18
#define DEFAULT_OSD_DURATION  5000
#define OSD_MESSAGE_LOW_PRIORITY    1
#define OSD_MESSAGE_NORMAL_PRIORITY 2

class KXineWidget : public virtual QWidget, private virtual QThread
{

 Q_OBJECT

public:
  KXineWidget(QWidget *parent=0, const char *name=0,
  		const QString& pathToConfigFile = QString::null, const QString& pathToLogoFile = QString::null,
		const QString& audioDriver = QString::null, const QString& videoDriver = QString::null,
		bool startManual = false, bool verbose = false);
 virtual ~KXineWidget();

  enum Speed { Normal = 0, Pause, Fast1, Fast2, Slow1, Slow2, Undefined };

  virtual bool initXine(); /* call this only if you set startManual = true in the constructor */
  virtual bool isXineReady() const { return m_xineReady; }
  virtual bool isPlaying() const;
  virtual void terminateThread();
  virtual Speed getSpeed() const { return m_currentSpeed; }
  virtual QString getXineLog() const;

  virtual void setQueue(const QStringList& urls) { m_queue = urls; }
  virtual void appendToQueue(const QString& url) { m_queue.append(url); } /* file to play; call slotPlay() to play queue... */
  virtual void clearQueue() { m_queue.clear(); }
  virtual bool isQueueEmpty() const { return m_queue.empty(); }

  virtual void showOSDMessage(const QString& message, uint duration = DEFAULT_OSD_DURATION /* ms */, int priority = OSD_MESSAGE_NORMAL_PRIORITY);

  /* stream info */
  virtual const QString& getURL() const { return m_trackURL; }
  virtual const QString& getTitle() const { return m_trackTitle; }
  virtual const QString& getArtist() const { return m_trackArtist; }
  virtual const QString& getAlbum() const { return m_trackAlbum; }
  virtual const QString& getTrackNumber() const { return m_trackNumber; } //may return an empty string with xine-lib 1.0
  virtual const QString& getGenre() const { return m_trackGenre; }
  virtual const QString& getYear() const { return m_trackYear; }
  virtual const QString& getComment() const { return m_trackComment; }
  virtual const QTime& getLength() const { return m_trackLength; }
  virtual const QString& getVideoCodec() const { return m_trackVideoCodec; }
  virtual QSize getVideoSize() const { return QSize(m_videoFrameWidth, m_videoFrameHeight); }
  virtual uint getVideoBitrate() const { return m_trackVideoBitrate; }
  virtual const QString& getAudioCodec() const { return m_trackAudioCodec; }
  virtual uint getAudioBitrate() const { return m_trackAudioBitrate; }
  virtual const QString& getSubtitleURL() const { return m_trackSubtitleURL; }
  virtual const QString& getSaveURL() const { return m_trackSaveURL; }
  virtual bool hasVideo() const { return m_trackHasVideo; }
  virtual bool hasAudio() const { return m_trackHasAudio; }
  virtual bool hasChapters() const { return m_trackHasChapters; }
  virtual bool hasSubtitleURL() const { return !m_trackSubtitleURL.isNull(); }
  virtual bool hasSaveURL() const { return !m_trackSaveURL.isNull(); }
  virtual bool isSeekable() const { return m_trackIsSeekable; }

  virtual uint getVolume() const; /* percent */
  virtual bool SoftwareMixing() const;
  virtual QTime getPlaytime() const;
  virtual uint getPosition() const; /* 0..65535 */

  virtual void savePosition(const int pos) { (pos > 0) ? m_savedPos = pos : m_savedPos = 0; }

  /* plugin handling */
  virtual QStringList getAudioDrivers() const { return m_audioDriverList; }
  virtual QStringList getVideoDrivers() const { return m_videoDriverList; }
  virtual void getAutoplayPlugins(QStringList& pluginNames) const;
  virtual bool getAutoplayPluginURLS(const QString& name, QStringList& urls);
  virtual QStringList getVisualPlugins() const;

#ifndef USE_QT_ONLY
  virtual void createDeinterlacePlugin(const QString& configString, QWidget* parent);
  virtual const QString getDeinterlaceConfig() const;
  virtual QStringList getVideoFilterNames() const;
  virtual QStringList getAudioFilterNames() const;
  virtual QStringList getAudioFilterConfig();
  virtual QStringList getVideoFilterConfig();
#endif

  virtual QString getSupportedExtensions() const;  /* get mime types xine can play */

  virtual void playNextChapter() const;
  virtual void playPreviousChapter() const;

  /* config stuff */
  virtual void setStreamSaveDir(const QString& dir);
  virtual const QString getStreamSaveDir();
  virtual void setBroadcasterPort(const uint port);
  virtual void getVideoSettings(int& hue, int& sat, int& contrast, int& bright,
                        int& avOffset, int& spuOffset) const;
 /* disc drives devices*/
  virtual QString audiocdDevice() const;
  virtual QString vcdDevice() const;
  virtual QString dvdDevice() const;

 /* take a screenshot */
  virtual QImage getScreenshot() const;
  virtual void getScreenshot(uchar*& rgb32BitData, int& videoWidth, int& videoHeight, double& scaleFactor) const;

 /* get the xine engine, needed by the xine configuration dialog */
  virtual const xine_t* const getXineEngine() const;
  virtual void saveXineConfig();

  /* for fullscreen */
  virtual void startMouseHideTimer();
  virtual void stopMouseHideTimer();

  virtual void globalPosChanged();  /* call this if global pos of parent was changed */

signals:
  virtual void signalXineFatal(const QString& message);   /** initXine() not successfull! **/
  virtual void signalXineError(const QString& message);
  virtual void signalXineMessage(const QString& message);
  virtual void signalXineStatus(const QString& status);
  virtual void signalXineReady();
  virtual void signalXinePlaying();

  virtual void signalHasChapters(bool);
  virtual void signalPlaybackFinished();
  virtual void signalNewChannels(const QStringList& audio, const QStringList& sub, int currentAudio, int currentSub);
  virtual void signalNewPosition(int, const QTime&);

  virtual void signalTitleChanged();
  virtual void signalVideoSizeChanged();
  virtual void signalLengthChanged();

  virtual void signalRightClick(const QPoint&);  /* right-click on video window */
  virtual void signalLeftClick(const QPoint&);
  virtual void signalMiddleClick();
  virtual void signalDoubleClick();
  virtual void signalSyncVolume();

public slots:
  virtual void polish();

  virtual void slotPlay();
  virtual void slotStop();

  virtual void slotSetVolume(int);
  virtual void slotToggleMute();
  virtual void slotSpeedPause();
  virtual void slotSpeedNormal();
  virtual void slotSpeedFaster();
  virtual void slotSpeedSlower();
  virtual void slotSetVisualPlugin(const QString& name);
  virtual void slotSetAudioChannel(int);
  virtual void slotSetSubtitleChannel(int);
  virtual void slotSetFileSubtitles(QString);
  virtual void slotStartSeeking();
  virtual void slotSeekToPosition(int pos);
  virtual void slotSeekToTime(const QTime&);
  virtual void slotStopSeeking();
  virtual void slotEnableVideoFilters(bool enable);
  virtual void slotEnableAudioFilters(bool enable);
  virtual void slotEject();
  virtual void slotEnableAutoresize(bool enable);
  virtual void slotAspectRatioAuto();
  virtual void slotAspectRatio4_3();
  virtual void slotAspectRatioAnamorphic(); /* 16:9 */
  virtual void slotAspectRatioSquare(); /* 1:1 */
  virtual void slotAspectRatioDVB(); /* 2.11:1 */
  virtual void slotZoomIn();
  virtual void slotZoomOut();
  virtual void slotZoomOff();
  virtual void slotToggleDeinterlace();
  virtual void slotSetDeinterlaceConfig(const QString&);
  void slotGetInfoDelayed();

  /* disc drive devices */
  virtual void slotSetAudiocdDevice(const QString&);
  virtual void slotSetVcdDevice(const QString&);
  virtual void slotSetDvdDevice(const QString&);

  virtual void slotSetHue(int hue);
  virtual void slotSetSaturation(int sat);
  virtual void slotSetContrast(int contrast);
  virtual void slotSetBrightness(int bright);
  virtual void slotSetAVOffset(int av);
  virtual void slotSetSpuOffset(int spu);

  virtual void slotSetEq30(int val);
  virtual void slotSetEq60(int val);
  virtual void slotSetEq125(int val);
  virtual void slotSetEq250(int val);
  virtual void slotSetEq500(int val);
  virtual void slotSetEq1k(int val);
  virtual void slotSetEq2k(int val);
  virtual void slotSetEq4k(int val);
  virtual void slotSetEq8k(int val);
  virtual void slotSetEq16k(int val);
  virtual void slotSetVolumeGain(bool gain);

 /* DVD Menus */
  virtual void slotMenuToggle();
  virtual void slotMenuTitle();
  virtual void slotMenuRoot();
  virtual void slotMenuSubpicture();
  virtual void slotMenuAudio();
  virtual void slotMenuAngle();
  virtual void slotMenuPart();
  virtual void slotDVDMenuLeft();
  virtual void slotDVDMenuRight();
  virtual void slotDVDMenuUp();
  virtual void slotDVDMenuDown();
  virtual void slotDVDMenuSelect();

 /***** postprocess filters ****/
  virtual void slotCreateVideoFilter(const QString& name, QWidget* parent);
  virtual void slotCreateAudioFilter(const QString& name, QWidget* parent);
  virtual void slotDeleteVideoFilter(PostFilter* filter);
  virtual void slotDeleteAudioFilter(PostFilter* filter);
  virtual void slotRemoveAllVideoFilters();
  virtual void slotRemoveAllAudioFilters();

protected slots:
  virtual void slotSendPosition();
  virtual void slotEmitLengthInfo();
  virtual void slotHideMouse();
  virtual void slotOSDHide();
  virtual void slotNoRecentMessage();

protected:
  virtual void mouseMoveEvent(QMouseEvent*);
  virtual void mousePressEvent(QMouseEvent*);
  virtual void timerEvent(QTimerEvent*);
  virtual void mouseDoubleClickEvent(QMouseEvent*);
  virtual void wheelEvent(QWheelEvent*);
  virtual void initOSD();

private:
  virtual void run();  /* implement pure virtual QThread method */

 /********* callbacks and threads ************/
  static void destSizeCallback(void* p, int video_width, int video_height, double video_aspect,
                       int* dest_width, int* dest_height, double* dest_aspect);

  static void frameOutputCallback(void* p, int video_width, int video_height, double video_aspect,
                          int* dest_x, int* dest_y, int* dest_width, int* dest_height,
                          double* dest_aspect, int* win_x, int* win_y);

  static void videoDriverChangedCallback(void* p, xine_cfg_entry_t* entry);
  static void audioDriverChangedCallback(void* p, xine_cfg_entry_t* entry);
  static void audioMixerMethodChangedCallback(void* p, xine_cfg_entry_t* entry);
  static void showOSDMessagesChangedCallback(void* p, xine_cfg_entry_t* entry);
  static void sizeForOSDMessagesChangedCallback(void* p, xine_cfg_entry_t* entry);
  static void fontForOSDMessagesChangedCallback(void* p, xine_cfg_entry_t* entry);

  static void xineEventListener(void* p, const xine_event_t*);

protected:
  virtual void sendXineError();
  virtual QTime getLengthInfo();
#ifndef USE_QT_ONLY
  virtual void wireVideoFilters();
  virtual void unwireVideoFilters();
  virtual void wireAudioFilters();
  virtual void unwireAudioFilters();
#endif

/* HELPER FUNCTIONS */
public:
#ifdef USE_QT_ONLY
  static QString i18n(const char *text);
#endif
  static void debugOut(QString);
  static void errorOut(QString);
  static void warningOut(QString);
  static QTime msToTime(int msec);
  static void   yuy2Toyv12(uint8_t *y, uint8_t *u, uint8_t *v, uint8_t *input, int width, int height);
  static uchar *yv12ToRgb(uint8_t *src_y, uint8_t *src_u, uint8_t *src_v, int width, int height);

protected:
  QStringList m_queue;

  bool m_startXineManual;
  bool m_xineReady;
  QString m_logoFile;
  QString m_preferedAudio;
  QString m_preferedVideo;
  bool m_xineVerbose;
  QString m_configFilePath;
  QStringList m_audioDriverList;
  QStringList m_videoDriverList;

 /*x11*/
  Window m_xineWindow;
  Display* m_xineDisplay;
  int m_xineScreen;

 /*xine*/
  x11_visual_t m_x11Visual;
  xine_t* m_xineEngine;
  xine_audio_port_t* m_audioDriver;
  xine_video_port_t* m_videoDriver;
  xine_stream_t* m_xineStream;
  xine_event_queue_t* m_eventQueue;
  xine_osd_t* m_osd;
  bool m_osdUnscaled;
  bool m_osdShow;
  int m_osdSize;
  char* m_osdFont;
  char** m_audioChoices;
  char* m_audioInfo;
  char** m_videoChoices;
  char* m_videoInfo;
  char* m_mixerInfo;
  char* m_osdShowInfo;
  char** m_osdSizeOptions;
  char* m_osdSizeInfo;
  char* m_osdFontInfo;

#ifndef USE_QT_ONLY
 /*postfilter*/
  QPtrList<PostFilter> m_videoFilterList;
  bool m_videoFiltersEnabled;
  QPtrList<PostFilter> m_audioFilterList;
  bool m_audioFiltersEnabled;
  PostFilter* m_deinterlaceFilter;
  bool m_deinterlaceEnabled;
  PostFilter* m_visualPlugin;
#else
  xine_post_t* m_xinePost;
  xine_post_out_t* m_postAudioSource;
  xine_post_in_t* m_postInput;
#endif

  QString m_audioDriverName;
  QString m_videoDriverName;
  QString m_visualPluginName;

  int m_videoFrameWidth;
  int m_videoFrameHeight;
  double m_videoAspect;
  int m_globalX;
  int m_globalY;

  Speed m_currentSpeed;
  QString m_xineMessage;
  QString m_xineError;
  QString m_statusString;
  bool m_softwareMixer;
  bool m_volumeGain;
  double m_displayRatio;
  QString m_cachedCDPath;
  QString m_cachedVCDPath;
  QString m_cachedDVDPath;
  QSize m_newParentSize;
  int m_currentZoom;
  QStringList m_audioCh;
  QStringList m_subCh;
  int m_currentAudio;
  int m_currentSub;
  int m_savedPos;
  bool m_autoresizeEnabled;
  bool m_DVDButtonEntered;
  QString m_newMRLReference;

  QTimer m_posTimer;
  QTimer m_lengthInfoTimer;
  uint m_lengthInfoTries;
  QTimer m_mouseHideTimer;
  QTimer m_osdTimer;
  QTimer m_recentMessagesTimer;

  QString m_trackURL;
  QString m_trackTitle;
  QString m_trackArtist;
  QString m_trackAlbum;
  QString m_trackNumber;
  QString m_trackGenre;
  QString m_trackYear;
  QString m_trackComment;
  QTime m_trackLength;
  QString m_trackVideoCodec;
  uint m_trackVideoBitrate;
  QString m_trackAudioCodec;
  uint m_trackAudioBitrate;
  QString m_trackSubtitleURL;
  QString m_trackSaveURL;
  bool m_trackHasVideo;
  bool m_trackHasAudio;
  bool m_trackHasChapters;
  bool m_trackIsSeekable;

/*  dvb  */
signals:
  virtual void stopDvb();

public:
  QString TimeShiftFilename;

public slots:
  virtual void setDvb( const QString &pipeName, const QString &chanName, int haveVideo );
  virtual void setDvbCurrentNext( const QStringList &list );
  virtual bool openDvb();
  virtual bool playDvb();
  virtual void slotPlayTimeShift();
  virtual void dvbShowOSD();
  virtual void dvbHideOSD();

private:
  int dvbHaveVideo;
  QTimer dvbOSDHideTimer;
  xine_osd_t *dvbOSD;
  QStringList dvbCurrentNext;
  unsigned int dvbColor[256];
  unsigned char dvbTrans[256];
  virtual unsigned int rgb2yuv( unsigned int R, unsigned int G, unsigned int B );
  virtual void initDvbPalette();
};



#endif
